/*
 * Copyright (C) 2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/*
 * Based on AudioCompress, (C) 2002-2004  M. Hari Nezumi <magenta@trikuare.cx>
 */


#include "para.h"
#include "compress.cmdline.h"

#define WAV_HEADER_LEN 44
#define GAINSHIFT 10 /* How fine-grained the gain is */

static int *peaks;
static struct gengetopt_args_info conf;

static void compress(void *data, unsigned long length)
{
	int16_t *audio = (int16_t *) data, *ap = audio;
	int peak = 1, pos = 0, i, gr, gf, gn;
	static int pn, clip, gainCurrent, gainTarget;

	/* determine peak's value and position */
	for (i = 0; i < length / 2; i++, ap++) {
		int val = ABS(*ap);
		if (val > peak) {
			peak = val;
			pos = i;
		}
	}
	peaks[pn] = peak;
	for (i = 0; i < conf.buckets_arg; i++) {
		if (peaks[i] > peak) {
			peak = peaks[i];
			pos = 0;
		}
	}
	/* determine target gain */
	gn = (1 << GAINSHIFT) * conf.target_level_arg / peak;
	if (gn < (1 << GAINSHIFT))
		gn = 1 << GAINSHIFT;
	gainTarget = (gainTarget * ((1 << conf.gain_smooth_arg) - 1) + gn)
		>> conf.gain_smooth_arg;
	/* give it an extra insignifigant nudge to counteract possible
	 * rounding error
	 */
	if (gn < gainTarget)
		gainTarget--;
	else if (gn > gainTarget)
		gainTarget++;
	if (gainTarget > conf.gain_max_arg << GAINSHIFT)
		gainTarget = conf.gain_max_arg << GAINSHIFT;
	/* see if a peak is going to clip */
	gn = (1 << GAINSHIFT) * 32768 / peak;
	if (gn < gainTarget) {
		gainTarget = gn;
		if (conf.anticlip_given)
			pos = 0;
	} else
		/* we're ramping up, so draw it out over the whole frame */
		pos = length;
	/* determine gain rate necessary to make target */
	if (!pos)
		pos = 1;
	gr = ((gainTarget - gainCurrent) << 16) / pos;
	gf = gainCurrent << 16;
	ap = audio;
	for (i = 0; i < length / 2; i++) {
		int sample;
		/* interpolate the gain */
		gainCurrent = gf >> 16;
		if (i < pos)
			gf += gr;
		else if (i == pos)
			gf = gainTarget << 16;
		/* amplify */
		sample = (*ap) * gainCurrent * conf.volume_arg / 10 >> GAINSHIFT;
		if (sample < -32768) {
			clip++;
			sample = -32768;
		} else if (sample > 32767) {
			clip++;
			sample = 32767;
		}
		*ap++ = sample;
	}
	pn = (pn + 1) % conf.buckets_arg;
//	fprintf(stderr, "bucket: %03i, peak: %05i, current gain: %03i, "
//		"clipped: %d\n", pn, peak, gainCurrent, clip);
}

int main(int argc, char *argv[])
{
	ssize_t len, ret;
	int16_t buf[4096];

	cmdline_parser(argc, argv, &conf);
	peaks = calloc(conf.buckets_arg, sizeof(int));
	len = read(STDIN_FILENO, buf, WAV_HEADER_LEN);
	if (len != WAV_HEADER_LEN) {
		len = -1;
		goto out;
	}
	len = write(STDOUT_FILENO, buf, WAV_HEADER_LEN);
	if (len != WAV_HEADER_LEN) {
		len = -1;
		goto out;
	}
//	fprintf(stderr, "wrote wave header, compresssing\n");
	while ((len = read(STDIN_FILENO, buf, 4096 * sizeof (int16_t))) > 0) {
		compress(buf, len);
		ret = write(STDOUT_FILENO, buf, len);
		if (ret == len)
			continue;
//		if (ret > 0)
//			fprintf(stderr, "compress: short write: %d/%d\n", ret, len);
		len = -1;
		break;
	}
out:
	return len < 0 ? 1 : 0;
}
